package sun.reflect.generics.reflectiveObjects;

import java.util.Arrays;
import sun.reflect.generics.visitor.Reifier;
import sun.reflect.generics.factory.GenericsFactory;
import sun.reflect.generics.tree.FieldTypeSignature;
import java.lang.reflect.Type;
import java.lang.reflect.WildcardType;

public class WildcardTypeImpl extends LazyReflectiveObjectGenerator implements WildcardType
{
    private Type[] upperBounds;
    private Type[] lowerBounds;
    private FieldTypeSignature[] upperBoundASTs;
    private FieldTypeSignature[] lowerBoundASTs;
    
    private WildcardTypeImpl(FieldTypeSignature[] upperBoundASTs, FieldTypeSignature[] lowerBoundASTs, GenericsFactory genericsFactory) {
        super(genericsFactory);
        this.upperBoundASTs = upperBoundASTs;
        this.lowerBoundASTs = lowerBoundASTs;
    }
    
    public static WildcardTypeImpl make(FieldTypeSignature[] array, FieldTypeSignature[] array2, GenericsFactory genericsFactory) {
        return new WildcardTypeImpl(array, array2, genericsFactory);
    }
    
    private FieldTypeSignature[] getUpperBoundASTs() {
        assert this.upperBounds == null;
        return this.upperBoundASTs;
    }
    
    private FieldTypeSignature[] getLowerBoundASTs() {
        assert this.lowerBounds == null;
        return this.lowerBoundASTs;
    }
    
    @Override
    public Type[] getUpperBounds() {
        if (this.upperBounds == null) {
            FieldTypeSignature[] upperBoundASTs = this.getUpperBoundASTs();
            Type[] upperBounds = new Type[upperBoundASTs.length];
            Reifier reifier;
            for (int i = 0; i < upperBoundASTs.length; upperBounds[i] = reifier.getResult(), ++i) {
                reifier = this.getReifier();
                upperBoundASTs[i].accept(reifier);
            }
            this.upperBounds = upperBounds;
        }
        return (Type[])this.upperBounds.clone();
    }
    
    @Override
    public Type[] getLowerBounds() {
        if (this.lowerBounds == null) {
            FieldTypeSignature[] lowerBoundASTs = this.getLowerBoundASTs();
            Type[] lowerBounds = new Type[lowerBoundASTs.length];
            Reifier reifier;
            for (int i = 0; i < lowerBoundASTs.length; lowerBounds[i] = reifier.getResult(), ++i) {
                reifier = this.getReifier();
                lowerBoundASTs[i].accept(reifier);
            }
            this.lowerBounds = lowerBounds;
        }
        return (Type[])this.lowerBounds.clone();
    }
    
    @Override
    public String toString() {
        Type[] lowerBounds = this.getLowerBounds();
        StringBuilder sb = new StringBuilder();
        if (lowerBounds.length > 0) {
            sb.append("? super ");
        }
        else {
            Type[] upperBounds = this.getUpperBounds();
            if (upperBounds.length > 0 && !upperBounds[0].equals(Object.class)) {
                lowerBounds = upperBounds;
                sb.append("? extends ");
            }
            else {
                return "?";
            }
        }
        assert lowerBounds.length > 0;
        int n = 1;
        for (Type type : lowerBounds) {
            if (n == 0) {
                sb.append(" & ");
            }
            n = 0;
            sb.append(type.getTypeName());
        }
        return sb.toString();
    }
    
    @Override
    public boolean equals(Object o) {
        if ((o instanceof WildcardType)) {
            WildcardType wildcardType = (WildcardType)o;
            return Arrays.equals(this.getLowerBounds(), wildcardType.getLowerBounds()) && Arrays.equals(this.getUpperBounds(), wildcardType.getUpperBounds());
        }
        return false;
    }
    
    @Override
    public int hashCode() {
        return Arrays.hashCode(this.getLowerBounds()) ^ Arrays.hashCode(this.getUpperBounds());
    }
}
